//////////////////////////////////////////////////////////////////
//
// file: Particles.cpp
// Description: Create particle objects from particles
//             in the tree. 
//
// Author: Reinhard Schwienhorst
// Created: 4/27/2004
//
// Modifications:
// 10/19/2006, RS: initial version
//
//
//////////////////////////////////////////////////////////////////
#include "single_top_analysis/Particles.hpp"
#include <iostream>
#include <string>
#include <fstream>

#include "single_top_analysis/utils.hpp"
//

using namespace std;

// needed to run with compiled Root:
ClassImp(Particles)


//-----------------------------------------------------------------------------------------
//
// Get the next event.
// It reads in the next event from the current chain and 
// Creates final state particles.
// This method returns -1 when there are no more events.
// It returns the total event count if successful.
int Particles::GetNextEvent()
{
  //cout<<"Particles::GetNextEvent, count "<<_eventCount<<endl;
  int countOffSet=0;
  // check that the chains are defined:
  if (_eventTree->fChain==0) return -1;


  // now Set things up
  int bytesread = _eventTree->fChain->GetEntry(_eventCount);
  
  if(bytesread==0) return -1;

  // Create 4-vectors
  readEvent();

  // If we are selecting by lepton, then check here for
  // the lepton flavor
  // 
  if(_leptonSelect>0) {
//     while( ! tll.hasAbsPdgId(_leptonSelect)) {
//       _eventCount++;
//       bytesread = _eventTree->fChain->GetEntry(_eventCount);
//       if(bytesread==0) return -1;
//       readEvent();
//     }    // end of while loop over events for checking lepton flavor
  } // end of: if only considering specific lepton flavor



  // Create final state objects like W bosons and top quarks
  //finalStatePass=CreateFinalState();
  
  // and finally increment the internal event counter
  _eventCount++;
  //cout<<"returning from GetNextEvent"<<endl;
  return(countOffSet+_eventCount-1);
} // end of GetNextEvent
  


/////////////////////////////////////////////////////////////////////
// constructor:
Particles::Particles():_eventTree(0)
		      , _eventCount(0),_tarGetTopMass(175.)
		      ,_leptonSelect(0),_normalization(1.)
{

}

////////////////////////////////////////////////////////////////////////
// destructor:
Particles::~Particles()
{
  if(_eventTree) delete _eventTree;
  _eventTree=0;
}

//////////////////////////////////////////////////////////////////////////
//
// initialize this class: read the chains in to Get the number of events.
void Particles::initialize(EV0* eventTree)
{

  // 
  // initialize the analysis classes
  _eventTree=eventTree;

  // now Set up the internal variables:
  _eventCount=0;
  electrons.clear();
  // 

  return;
}


//
// make particles into 4-vectors for easier manipulation
int Particles::readEvent()
{
  // Set the event weight if there is any
  eventWeight = 1.;  //_eventTree->eventWeight;
  runNumber   = _eventTree->runNumber;
  eventNumber = _eventTree->eventNumber;

  // Set weight from command line
  eventWeight *= _normalization;

  // reSet all of the particles
  electrons.clear();




  // now start filling
  //
  // begin with electrons
  Electron newEle;

  for(int iE=0;iE<_eventTree->El_N;iE++) {
    Double_t elPt  =_eventTree->El_p_T->operator[](iE);
    Double_t elEta =_eventTree->El_eta->operator[](iE);
    Double_t elPhi =_eventTree->El_phi->operator[](iE);
    Double_t elE   =_eventTree->El_E->operator[](iE);
    //Double_t elEoverP=_eventTree->El_EoverP->operator[](iE);
    Double_t elCharge=_eventTree->El_charge->operator[](iE);

    newEle.SetPtEtaPhiE(elPt,elEta,elPhi,elE);
    newEle.SetCharge(elCharge);
    //newEle.SetEoverP(elEoverP);
    electrons.push_back(newEle);
  }

  return 0;
}


//-------------------------------------------------------------
// Create final state particles and make cuts if necessary (require at least
// two jets and at least one isolated lepton)
// return false if these conditions are not fulfilled
bool Particles::CreateFinalState()
{
//   //cout<<"begin of particles::Create final state"<<endl;

  //finalStatePass=true;
  
//   //
//   // for the neutrino we pretend that we don't know where it is.
//   // All we can do is reconstruct it.
//   Double_t nu_px = -tll.Px();
//   Double_t nu_py = -tll.Py();
//   for(int i=0;i<njets;i++) {
//     nu_px += -finalJets[i].Px();
//     nu_py += -finalJets[i].Py();
//   }
//   // last argument to neutrino_pz is a large negative number
//   // to indicate that we don't want to use a scale factor.
//   Double_t scf=-99999.;
//   //cout<<"calling pzneutrino"<<endl;
//   Double_t nu_pz=Neutrino_Pz(nu_px,nu_py,
// 			  tll.Px(),tll.Py(),tll.Pz(),tll.E(),scf,tlnu.Pz());
//   //

//   // in the D0 analysis we use the scale factor, but here we don't
//   // because we know neutrino px and py very well.
//   if(fabs(nu_pz)>tlnu.Pz()) {
//     //cout<<" making neutrino, initial PZ is "<<tlnu.Pz()<<", final Pz "<<nu_pz<<endl;
//   }
//   Double_t nu_e=sqrt(nu_px*nu_px + nu_py*nu_py + nu_pz*nu_pz);
//   finalNeutrino.SetPxPyPzE(nu_px,nu_py,nu_pz,nu_e);
//   finalNeutrino.SetPdgId(-12);
  

//   // now reconstruct the W and the top quark
//   //cout<<"making W"<<endl;
//   finalW = finalNeutrino;
//   finalW += finalLepton;
//   //cout<<"final W p "<<finalW.P()<<", E "<<finalW.E()<<", M "<<finalW.M()<<endl;

//   //cout<<"Assigning jets to b and bbar quarks"<<endl;
//   // and assign the final jets to the b and bbar quark
//   bJetPtr=0;
//   bbarJetPtr=0;
//   gluonJetPtr=0;
//   lqJetPtr=0;
//   for(int i=0;i<njets;i++) {
//     if(finalJets[i].hasPdgId(5)) {
//       // this is the b quark jet
//       bJetPtr=&(finalJets[i]);
//       //cout<<"Jet "<<i<<" is the b."<<endl;
//     }
//     if(finalJets[i].hasPdgId(-5)) {
//       // this is the bbar quark jet
//       bbarJetPtr=&(finalJets[i]);
//       //cout<<"Jet "<<i<<" is the bbar."<<endl;
//     }
//     if(finalJets[i].hasPdgId(21)) {
//       // this is the gluon jet
//       gluonJetPtr=&(finalJets[i]);
//       //cout<<"Jet "<<i<<" is the gluon."<<endl;
//     }
//     if(finalJets[i].hasPdgId(1)) {
//       // this is the light quark jet
//       lqJetPtr=&(finalJets[i]);
//       //cout<<"Jet "<<i<<" is the light quark."<<endl;
//     }
//   } // end of loop over jets
//   //
//   // we are not yet re-creating the jets from the top quark here,
//   // instead we are waiting for the user to do so explicitly.
//   // simply reSet all of the pointers at this point.
//   finalTop_leadingJet.SetPxPyPzE(0.,0.,0.,0.);
//   finalTop_bestJet.SetPxPyPzE(0.,0.,0.,0.);
//   finalTop_bestJet12.SetPxPyPzE(0.,0.,0.,0.);
//   finalTop_bgJet.SetPxPyPzE(0.,0.,0.,0.);
//   bestJet12Ptr=0;
//   nonBestJet12Ptr=0;
//   bestJetPtr=0;
//   bgJet.SetPxPyPzE(0.,0.,0.,0.);
//   bgJetbJetPtr=0;
//   nonBgJetPtr=0;
//   bgJetGluonPtr=0;
//   taggedJetPtr=0;
//   untaggedPtJetPtr=0;
//   untaggedEtaJetPtr=0;
//   forwardJetPtr=0;
  
  //cout<<"end of particles::Create final state"<<endl;
  //return finalStatePass;
  return true;
} 



//--------------------------------------------------------------------
// make top quarks
//
//
void Particles::makeTopQuarks()
{
  //
  // make top quarks from the W and several possible jets. 
//   // here we can Create several top quarks
//   // start by making one from the w and the leading jet
//   finalTop_leadingJet = finalW;
//   finalTop_leadingJet += finalJets[0];
  
  
//   //cout<<"finding best jet"<<endl;
//   // now use the best jet (the one that reconstructs to a
//   // top mass closest to the tarGet top mass (default=175GeV)
//   TLorentzVector tempJets[4];
//   for(int iJet=0;iJet<njets;iJet++) tempJets[iJet]=finalJets[iJet];

//   int ibestjet = findBestJet(njets,tempJets,finalW,_tarGetTopMass);
//   bestJetPtr=&(finalJets[ibestjet]);
//   finalTop_bestJet = finalW;
//   finalTop_bestJet += *bestJetPtr;
//   // for bestjet12 only consider leading two jets
//   int itemp=njets>2? 2: njets;
//   ibestjet = findBestJet(itemp,tempJets,finalW,_tarGetTopMass);
//   bestJet12Ptr=&(finalJets[ibestjet]);
//   finalTop_bestJet12 = finalW;
//   finalTop_bestJet12 += *(bestJet12Ptr);
  
//   // the non-best jet
//   if(njets>=2) {
//     nonBestJet12Ptr=&(finalJets[1-ibestjet]);
//   } else {
//     nonBestJet12Ptr=0;
//   }

//   //////////////////////////////////////////////////////////////////////
//   //
//   //cout<<"add in a possible gluon"<<endl;
//   // add in the gluon if it is close to this jet
//   // and then determine the best jet
//   if(njets<=2) {
//     // nothing to be done, bestJet12 is the same as this
//     finalTop_bgJet=finalTop_bestJet12;
//     bgJet=*bestJet12Ptr;
//     bgJetbJetPtr=bestJet12Ptr;
//     nonBgJetPtr=nonBestJet12Ptr;
//   }
//   else {
//     // there are three jets in the event, pair each of the leading two
//     // with the gluon and then find the best jet
//     for(int i=0;i<2;i++) {
//       tempJets[i]=finalJets[i];
//       tempJets[i+2] = finalJets[i];
//       tempJets[i+2] += finalJets[2];
//     } // end of loop over jets 1,2
//     //cout<<"finding best jet out of 4"<<endl;
//     ibestjet=findBestJet(4,tempJets,finalW,_tarGetTopMass);
//     finalTop_bgJet=tempJets[ibestjet];
//     finalTop_bgJet += finalW;
//     bgJet=tempJets[ibestjet];
//     // figure out if we included the gluon or not.
//     if(ibestjet>=2) {
//       bgJetbJetPtr=&(finalJets[ibestjet-2]);
//       nonBgJetPtr=&(finalJets[1-(ibestjet-2)]);
//       bgJetGluonPtr=&(finalJets[2]);
//     }
//     else {
//       bgJetbJetPtr=&(finalJets[ibestjet]);
//       nonBgJetPtr=&(finalJets[1-(ibestjet)]);
//       bgJetGluonPtr=0;
//     }
//   } // end of else: try to include the gluon

//   // final top from tagged jet and W
//   if(taggedJetPtr!=0) {
//     finalTop_taggedJet=*taggedJetPtr;
//     finalTop_taggedJet += finalW;
    
//     // and the W using a top mass constraint:
//     // the neutrino from the top mass constraint
//     // also use a constraint on the top mass rather than the smaller of the two
//     // solutions.
//     //cout<<"True Top mass is "<<tlt.M()<<" e= "<<tlt.E()<<", px= "<<tlt.Px()<<", py= "<<tlt.Py()<<", pz= "<<tlt.Pz()<<endl;
//     Double_t nu_pz2=Neutrino_Pz_Top(finalNeutrino.Px(),finalNeutrino.Py(),
// 				  finalLepton.Px(),finalLepton.Py(),finalLepton.Pz(),finalLepton.E(),
// 				  taggedJetPtr->Px(),taggedJetPtr->Py(),taggedJetPtr->Pz(),taggedJetPtr->E(),
// 				  tlnu.Pz());
    
    
//     ////
//     Double_t nu_e=sqrt(pow(finalNeutrino.Px(),2) + pow(finalNeutrino.Py(),2) + nu_pz2*nu_pz2);
//     finalNeutrinoTopC.SetPxPyPzE(finalNeutrino.Px(),finalNeutrino.Py(),nu_pz2,nu_e);
//     finalNeutrinoTopC.SetPdgId(-12);
    
//     // the final lepton from top constraint on neutrino
//     finalWTopC = finalNeutrinoTopC;
//     finalWTopC += finalLepton;

//     finalTop_taggedJetTopC =*taggedJetPtr;
//     finalTop_taggedJetTopC += finalWTopC;
//   }
//   else {
//     finalTop_taggedJet.SetPxPyPzE(0.,0.,0.,0.);
//     finalTop_taggedJetTopC.SetPxPyPzE(0.,0.,0.,0.);
//   }


//   //Particle tempt=finalW;
//   //tempt += bJet;
//   //cout<<"top mass "<<tempt.M()<<", "<<finalTop_bestJet12.M();
//   //cout<<" final jets: "<<(finalW+finalJets[0]).M()<<", "<<(finalW+finalJets[1]).M()<<endl;
//   //cout<<"best jet pt, phi is "<<bestJet12.Pt()<<", "<<bestJet12.Phi()<<endl;
//   //cout<<"   b jet pt, phi is "<<bJet.Pt()<<", "<<bJet.Phi()<<endl;


  return;
}

