////////////////////////////////////////////////////////////////////////////////////////////
//
// Helper functions that could be useful.
//
//
//
// Author: Reinhard Schwienhorst
//
// Modifications:
// 9/2004, RS: Add method that determines neutrino pz with top mass constraint
//
//
////////////////////////////////////////////////////////////////////////////////////////////
#include "TMath.h"
#include "single_top_analysis/utils.hpp"
#include <string>
#include <iostream>

using namespace std;

// convert detector eta into Physics eta.
// R is the distance from beamline in cm(~72 is CPS)
// This simply calls the method below with -zvertex
Double_t etaDetector2Physics(const Double_t& etain,const Double_t& zvertex,const Double_t& R)
{
  return(etaPhysics2Detector(etain,-zvertex,R));
}

// convert Physics eta into detector eta.
// This one was written by Adam Yurkewicz.
Double_t etaPhysics2Detector(const Double_t& physEta, const Double_t& zvertex,const Double_t& R) {

  Double_t detEta=0.;
  //do case where physEta is positive
  if (physEta>=0.) {
    //physEta is positive, zvertex is positive
    if (zvertex>=0) {
      Double_t Theta=0.;
      Double_t ThetaPrime=0.;
      Theta = 2.*TMath::ATan(exp(-1.*physEta));
      ThetaPrime = TMath::ATan(R/(zvertex+R/TMath::Tan(Theta)));
      detEta = -1.*TMath::Log(TMath::Tan(ThetaPrime/2.));
    }
    //physEta is positive, zvertex is negative
    else {
      Double_t Theta=0.;
      Double_t ThetaPrime=0.;
      Theta = 2.*TMath::ATan(exp(-1.*physEta));
      Double_t d = R/TMath::Tan(Theta);
      if (d>(-1.*zvertex)) {
        ThetaPrime = TMath::ATan(R/(zvertex+d));
        detEta = -1.*TMath::Log(TMath::Tan(ThetaPrime/2.));
      }
      //case where positive physEta becomes negative detector Eta
      else {
        ThetaPrime = TMath::ATan(R/(-1.*(zvertex+d)));
        detEta = TMath::Log(TMath::Tan(ThetaPrime/2.));
      }
    }
  }
  //now do case where physEta is negative
  else {
    //physEta is negative, zvertex is positive
    if (zvertex>=0) {
      Double_t Theta=0.;
      Double_t ThetaPrime=0.;
      Theta = 2.*TMath::ATan(exp(physEta));
      //      cout<<Theta<<endl;
      Double_t d = R/TMath::Tan(Theta);
      if (d>zvertex) {
        ThetaPrime = TMath::ATan(R/(d-zvertex));
        //      cout<<ThetaPrime<<endl;
        detEta = TMath::Log(TMath::Tan(ThetaPrime/2.));
      }
      //case where negative physEta becomes positive detector Eta
      else {
        ThetaPrime = TMath::ATan(R/(zvertex-d));
        detEta = -1.*TMath::Log(TMath::Tan(ThetaPrime/2.));
      }
    }
    //physEta is negative, zvertex is negative
    else {
      Double_t Theta=0.;
      Double_t ThetaPrime=0.;
      Theta = 2.*TMath::ATan(exp(physEta));
      ThetaPrime = TMath::ATan(R/(-1.*zvertex+R/TMath::Tan(Theta)));
      detEta = TMath::Log(TMath::Tan(ThetaPrime/2.));
    }
  }
  return detEta;
}

///////////////////////////////////////////////////////////////////////////
// Get the neutrino PZ from the W mass constraint.
// pick the solution with the smaller pz.  
// this function is from the D0 top_tree package.
// Modification: only do the scaling if scf is passed in (i.e. if it is not -99999.).
Double_t Neutrino_Pz( Double_t nu_px, Double_t nu_py, 
		    Double_t l_px, Double_t l_py, Double_t l_pz, Double_t l_e, 
		    Double_t& scf,Double_t nu_pz_true)
{
  Double_t solution1;
  Double_t solution2;
  // find the two possible solutions:
  Neutrino_Pz_Base( nu_px, nu_py, l_px, l_py, l_pz, l_e, // input arguments 
		    scf, solution1, solution2);          // return variables
  // now choose the lower one of the two
  //if(nu_pz_true>100)
  //cout<<"possible solutions: "<<solution1<<" and "<<solution2;
  //cout<<" and truth is "<<nu_pz_true<<" and pz_l "<<l_pz<<endl;
  // choose solution with smallest |l_pz| a la Run I
  // alternatives don't work !!
  
    Double_t nu_pz = fabs (solution1) < fabs (solution2) ? solution1 : solution2 ;

  
    //cout<<"chose "<<nu_pz<<endl;
    return nu_pz;
  
}
 
///////////////////////////////////////////////////////////////////////////
// Get the neutrino PZ from a W mass constraint.
// pick the solution giving a top mass closest to 178GeV
// 
Double_t Neutrino_Pz_Top( Double_t nu_px, Double_t nu_py, 
			Double_t l_px, Double_t l_py, Double_t l_pz, Double_t l_e, 
			Double_t b_px, Double_t b_py, Double_t b_pz, Double_t b_e, 
			Double_t nu_pz_true)
{
  Double_t solution1;
  Double_t solution2;
  Double_t scf=-99999.;

  // find the two possible solutions:
  Neutrino_Pz_Base( nu_px, nu_py, l_px, l_py, l_pz, l_e,  // input arguments 
		    scf, solution1, solution2);           // return variables
  // now choose a solution
  //cout<<"  possible solutions: "<<solution1<<" and "<<solution2;
  //cout<<" and truth is "<<nu_pz_true<<endl;
  
  // compute the two solutions to the top mass:

  // compute two W invariant masses:
  Double_t nu_e1=sqrt(pow(nu_px,2)+pow(nu_py,2)+pow(solution1,2));
  //Double_t w_m1=sqrt( pow(nu_e1+l_e,2)-
  //	    (pow(nu_px+l_px,2)+pow(nu_py+l_py,2)+pow(solution1+l_pz,2)));
  Double_t nu_e2=sqrt(pow(nu_px,2)+pow(nu_py,2)+pow(solution2,2));
  //Double_t w_m2=sqrt( pow(nu_e2+l_e,2)-
  //	    (pow(nu_px+l_px,2)+pow(nu_py+l_py,2)+pow(solution2+l_pz,2)));

  Double_t bl_px=l_px+b_px;
  Double_t bl_py=l_py+b_py;
  Double_t bl_pz=l_pz+b_pz;
  Double_t bl_e =l_e +b_e;
  Double_t top_e1=bl_e+nu_e1;
  Double_t top_e2=bl_e+nu_e2;
  Double_t top_px=bl_px+nu_px;
  Double_t top_py=bl_py+nu_py;
  Double_t top_pz1=bl_pz+solution1;
  Double_t top_m1=sqrt(pow(top_e1,2)-pow(top_px,2)-pow(top_py,2)-pow(top_pz1,2));
  Double_t top_pz2=bl_pz+solution2;
  Double_t top_m2=sqrt(pow(top_e2,2)-pow(top_px,2)-pow(top_py,2)-pow(top_pz2,2));
  //cout<<" reco Top mass 1 is "<<top_m1<<", e1= "<<top_e1<<", pz1= "<<top_pz1<<endl;
  //cout<<" reco Top mass 2 is "<<top_m2<<", e2= "<<top_e2<<", pz2= "<<top_pz2<<endl;
  
  Double_t nu_pz = fabs (top_m1-178.) < fabs (top_m2-178.) ? solution1 : solution2 ;
  
  
  //cout<<"  choose "<<nu_pz<<endl;
  return nu_pz;
  
}
 
///////////////////////////////////////////////////////////////////////////
// Get the neutrino PZ from the W mass constraint.
// This method is the basic method that calculates the two  
// possible solutions.
//
// This method assumes that the lepton and the neutrino are massless.
//
void Neutrino_Pz_Base( Double_t nu_px, Double_t nu_py, 
		       Double_t l_px, Double_t l_py, Double_t l_pz, Double_t l_e,
		       Double_t& scf,Double_t &solution1, Double_t &solution2)
{
  bool usescf=true;
  if(scf<0.) {
    usescf=false;
    //cout<<"not using scale factor!"<<endl;
  }
  scf = 1.0;
  Double_t nu_pz = 0.;
  Double_t l_pt  = TMath::Sqrt(pow(l_px,2)+pow(l_py,2));
  
  Double_t Mw    = 80.43;
  Double_t Mt;
  Double_t nu_pt = TMath::Sqrt(pow(nu_px,2)+pow(nu_py,2));
  Mt = TMath::Sqrt(pow(l_pt+nu_pt,2)-
		   pow(l_px+nu_px,2)-
		   pow(l_py+nu_py,2));
  
  //cout<<"   W mass is "<<Mw<<", mt is "<<Mt<<endl;
  Double_t A;
  //rs: only change the scale factor if the user wants to.
  if (Mt<Mw || !usescf) A = pow(Mw,2)/2.;
  else       {           // assume Mt=Mw, rescale MET accordingly
    //cout<<"  scaling met"<<endl;
    A = pow(Mt,2)/2.;
    Double_t k = nu_pt*l_pt - nu_px*l_px - nu_py*l_py;
    k = (k == 0 ? 0.00001 : k);
    scf = 0.5*pow(Mw,2)/k ;
    nu_px *= scf;
    nu_py *= scf;
    nu_pt = sqrt(pow(nu_px,2)+pow(nu_py,2));
  }    
  
  // if the electron goes along the beam direction (Px==Py==0),
  // then the algorithm below doesn't work, so do the calculation here.
  Double_t S_denom=pow(l_pz,2)-pow(l_e,2);
  if(S_denom==0.) {
    //cout<<"lepton along beam direction, Mt is "<<Mt<<endl;
    //cout<<"lpx "<<l_px<<", l_py "<<l_py<<", l_pz "<<l_pz<<", n_px "<<nu_px<<", n_py "<<nu_py<<endl;
    // in this case there is one exact solution
    Double_t mwplz=pow(Mw,2)/fabs(l_pz);
    nu_pz= (pow(nu_px,2) + pow(nu_py,2) - pow(mwplz/2.,2))/mwplz;
    solution1=nu_pz;
    solution2=nu_pz;
    return;
  }
  
  Double_t B = nu_px*l_px + nu_py*l_py;
  Double_t C;
  Double_t C1 = 1.0E0+pow(nu_pt,2)*(pow(l_pz,2)-pow(l_e,2))/pow(A+B,2);
  if(C1<=0.) {
    //cout<<"Warning in Neutrino_Pz! negative C1 "<<C1<<", Setting it to 0."<<endl;
    // in this case C will be zero and there is only one solution, 
    // so simply return the one possible solution
    solution1= (-(A+B)*l_pz) / (pow(l_pz,2)-pow(l_e,2));
    solution2=solution1;
    return;
  }
  else {
    C = TMath::Sqrt(C1);
    solution1= (-(A+B)*l_pz + (A+B)*l_e*C) / S_denom;
    solution2= (-(A+B)*l_pz - (A+B)*l_e*C) / S_denom;
    return;
  }
  
}

//
// find the "best jet", which is the one that gives a top mass
// closest to 175GeV.
// return the index in the array that points to the best jet.
int findBestJet(int njets,TLorentzVector* jets,TLorentzVector W,Double_t tarGetmass)
{
  // loop over the jets and for each reconstruct the invariant mass
  // of the jet-W system.
  TLorentzVector top;
  Double_t mt,diff,mindiff=999999999.;
  Double_t mtarg=pow(tarGetmass,2);
  //
  int foundjet=0;
  //cout<<"in findBestJet there are njets "<<njets<<endl;
  for(int i=0;i<njets;i++) {
    top = W + jets[i];
    // for some reason M() returns a negative mass every other time,
    // so we have to make sure we take the absolute value
    mt=top.Mag2();
    diff=fabs(mt-mtarg);
    //cout<<"  jet "<<i<<", mass "<<top.M()<<", diff "<<diff<<endl;
    if(diff<mindiff) {
      mindiff=diff;
      foundjet=i;
      //cout<<"  best mass for jet "<<i<<" is "<<top.M()<<endl;
    }
  } // end of loop over jets

  return foundjet;
}
